from extractor import extractor as ext

import os
import numpy
import zfit
import math
import pprint
import pytest
import utils_noroot as utnr
import rk.utilities as rkut

from logzero    import logger    as log
from np_reader  import np_reader as np_rdr
from rkex_model import model

#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return
#----------------------------------------------------
def check_close(value, target, abs_tol=1e-4):
    pas_val = math.isclose(value, target, abs_tol = abs_tol)
    if not pas_val:
        log.error(f'{value:.6f} != {target:.6f}')
        raise
#----------------------------------------------------
def test_simple():
    log.info('Running: test_simple')
    d_eff       = model.get_eff()
    d_yld       = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 

    mod         = model(preffix='simple', d_eff=d_eff)
    d_dat       = mod.get_data(d_nent=d_yld)
    d_mod       = mod.get_model()

    obj         = ext()
    obj.eff     = d_eff 
    obj.data    = d_dat
    obj.model   = d_mod 
    obj.plt_dir = 'tests/extractor/simple'

    result      = obj.get_fit_result()
    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']
#----------------------------------------------------
def test_efficiency():
    log.info('Running: test_efficiency')
    d_eff        = model.get_eff(kind='diff')
    d_yld        = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 

    mod          = model(preffix='efficiency', d_eff=d_eff)
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext()
    obj.eff      = d_eff 
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/efficiency'

    result       = obj.get_fit_result()
    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']
#----------------------------------------------------
def test_constraint():
    log.info('Running: test_constraint')
    d_eff        = model.get_eff(kind='half')
    d_yld        = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 

    mod          = model(preffix='constraint', d_eff=d_eff)
    cvmat        = mod.get_cov(kind='diag_eq', c=0.001)
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext()
    obj.eff      = d_eff 
    obj.cov      = cvmat
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/constraint'

    result       = obj.get_fit_result()

    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']
#----------------------------------------------------
def test_rjpsi():
    log.info('Running: test_rjpsi')
    d_eff        = model.get_eff(kind='half')
    d_yld        = {'d1' : 2e3, 'd2' : 2e3, 'd3' : 2e3, 'd4' : 2e3} 

    mod          = model(preffix='rjpsi', d_eff=d_eff)
    cvmat        = mod.get_cov(kind='diag_eq', c=0.001)
    d_rjpsi      = mod.get_rjpsi(kind='eff_bias')
    d_dat        = mod.get_data(d_nent=d_yld)
    d_mod        = mod.get_model()

    obj          = ext()
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cvmat
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/rjpsi'

    result       = obj.get_fit_result()

    result.hesse()

    d_rk   = result.params['rk']
    rk_val = d_rk['value']
    rk_err = d_rk['hesse']['error']
#----------------------------------------------------
def test_real():
    skip_test()
    log.info('Running: test_real')

    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    cv_sys       = rdr.get_cov(kind='sys')
    cv_sta       = rdr.get_cov(kind='sta')
    d_eff        = rdr.get_eff()
    d_rjpsi      = rdr.get_rjpsi()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    mod          = model(preffix='real', d_eff=d_eff)
    d_mod        = mod.get_model()
    d_dat        = mod.get_data(d_nent=d_rare_yld)

    obj          = ext()
    obj.rjpsi    = d_rjpsi
    obj.eff      = d_eff
    obj.cov      = cv_sys + cv_sta
    obj.data     = d_dat
    obj.model    = d_mod 
    obj.plt_dir  = 'tests/extractor/real'
    result       = obj.get_fit_result()

    log.info(f'Calculating errors')
    result.hesse()
    result.freeze()
    utnr.dump_pickle(result, 'tests/extractor/real/result.pkl')
#----------------------------------------------------
def main():
    test_real()
    test_rjpsi()
    test_constraint()
    test_simple()
    test_efficiency()
#----------------------------------------------------
if __name__ == '__main__':
    main()




import numpy
import zfit
import os
import matplotlib.pyplot as plt

from logzero     import logger  as log
from zutils.plot import plot    as zfp
from scipy.stats import poisson

#----------------------------------------------------
class model:
    def __init__(self, rk=1, preffix='', d_eff=None):
        self._obs    = zfit.Space('x', limits=(0, 100))
        self._rk     = rk
        self._preffix= preffix
        self._d_eff  = d_eff 
        self._out_dir= None

        zfit.settings.changed_warnings.hesse_name = False

        self._d_mod = None
    #----------------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create: {value}')
            raise

        self._out_dir = value
        log.debug(f'Using output directory: {self._out_dir}')
    #----------------------------------------------------
    def _get_data_simple(self, nentries):
        nentries = int(nentries)

        arr_sig = numpy.random.normal(size=nentries, loc=50, scale=3)
        arr_bkg = numpy.random.exponential(size=nentries, scale=100)
    
        arr_dat = numpy.concatenate([arr_sig, arr_bkg])
    
        return arr_dat
    #----------------------------------------------------
    def _get_gauss(self, preffix=''):
        preffix= f'{preffix}_{self._preffix}'

        mu     = zfit.Parameter(f'mu_{preffix}', 50., 40,  60)
        sg     = zfit.Parameter(f'sg_{preffix}', 2.0,  1,   5)
        gauss  = zfit.pdf.Gauss(obs=self._obs, mu=mu, sigma=sg)
        nsg    = zfit.Parameter(f'nsg_{preffix}', 10, 0, 100000)
        esig   = gauss.create_extended(nsg)
    
        lb     = zfit.Parameter(f'lb_{preffix}', -0.05,  -0.10, 0.00)
        exp    = zfit.pdf.Exponential(obs=self._obs, lam=lb)
        nbk    = zfit.Parameter(f'nbk_{preffix}', 10, 0, 100000)
        ebkg   = exp.create_extended(nbk)
    
        pdf    = zfit.pdf.SumPDF([esig, ebkg]) 
    
        return pdf 
    #----------------------------------------------------
    def _get_ds_model(self, ds):
        self._pdf_mm = self._get_gauss(preffix=f'mm_{ds}')
        self._pdf_ee = self._get_gauss(preffix=f'ee_{ds}')
    
        return self._pdf_mm, self._pdf_ee
    #----------------------------------------------------
    def _get_ds_data(self, nentries, eff_ee=1.0, eff_mm=1.0):
        nentries_mm = poisson.rvs(nentries, size=1)[0]
        nentries_ee = poisson.rvs(nentries, size=1)[0]

        arr_mm      = self._get_data_simple(self._rk * nentries_mm)
        arr_ee      = self._get_data_simple(           nentries_ee)

        arr_flg_mm  = numpy.random.binomial(1, eff_mm, arr_mm.shape[0]) == 1
        arr_flg_ee  = numpy.random.binomial(1, eff_ee, arr_ee.shape[0]) == 1
    
        arr_mm      = arr_mm[arr_flg_mm] 
        arr_ee      = arr_ee[arr_flg_ee] 
        
        dst_mm = zfit.Data.from_numpy(obs=self._obs, array=arr_mm)
        dst_ee = zfit.Data.from_numpy(obs=self._obs, array=arr_ee)
    
        return dst_mm, dst_ee
    #----------------------------------------------------
    def get_data(self, d_nent=None, rseed=3):
        if self._d_eff is None:
            log.error(f'No efficiencies found, cannot provide data')
            raise

        numpy.random.seed(seed=rseed)

        d_data     = {}
        dst_mm_tos = None
        for ds, (eff_mm, eff_ee) in self._d_eff.items():
            ds_only    = ds.split('_')[0]
            nentries   = d_nent[ds_only]
            log.debug(f'Dataset: {ds}[{nentries}]')

            dst_mm, dst_ee = self._get_ds_data(nentries, eff_ee=eff_ee, eff_mm=eff_mm)
            if 'TIS' in ds:
                dst_mm     = dst_mm_tos
            else:
                dst_mm_tos = dst_mm

            log.debug(f'Electron data: {dst_ee.numpy().shape[0]}')
            log.debug(f'Muon data: {dst_mm.numpy().shape[0]}')

            d_data[ds]     = dst_mm, dst_ee

        for key, (dat_mm, dat_ee) in d_data.items():
            self._plot_data(f'{key}_mm', dat_mm)
            self._plot_data(f'{key}_ee', dat_ee)
    
        return d_data
    #----------------------------------------------------
    def get_cov(self, kind='diag', c = 0.01):
        l_mat = []
        if kind == 'diag_eq':
            l_mat = [[ c, 0, 0, 0 ],
                     [ 0, c, 0, 0 ],
                     [ 0, 0, c, 0 ],
                     [ 0, 0, 0, c ]]
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return numpy.array(l_mat)
    #----------------------------------------------------
    def get_rjpsi(self, kind='one'):
        d_rjpsi = {}
    
        if   kind == 'one':
            d_rjpsi['d1'] = 1 
            d_rjpsi['d2'] = 1 
            d_rjpsi['d3'] = 1 
            d_rjpsi['d4'] = 1 
        elif kind == 'eff_bias':
            d_rjpsi['d1'] = 0.83333333 
            d_rjpsi['d2'] = 0.83333333 
            d_rjpsi['d3'] = 0.83333333 
            d_rjpsi['d4'] = 0.83333333 
        else:
            log.error(f'Wrong kind: {kind}')
            raise
    
        return d_rjpsi
    #----------------------------------------------------
    def _plot_model(self, key, mod):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots/models'
        os.makedirs(plt_dir, exist_ok=True)

        obj= zfp(data=mod.create_sampler(n=10000), model=mod)
        obj.plot(nbins=50) 

        log.info(f'Saving to: {plt_dir}/{key}.png')
        plt.savefig(f'{plt_dir}/{key}.png')
        plt.close('all')
    #----------------------------------------------------
    def _plot_data(self, key, dat):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots/data'
        os.makedirs(plt_dir, exist_ok=True)

        arr_dat = dat.value().numpy()

        plt.hist(arr_dat, bins=50)

        log.info(f'Saving to: {plt_dir}/{key}.png')
        plt.savefig(f'{plt_dir}/{key}.png')
        plt.close('all')
    #----------------------------------------------------
    def get_model(self):
        if self._d_mod is not None:
            return self._d_mod
    
        d_mod       = {}
        if self._d_eff is None:
            d_mod['d1'] = self._get_ds_model('d1')
            d_mod['d2'] = self._get_ds_model('d2')
            d_mod['d3'] = self._get_ds_model('d3')
            d_mod['d4'] = self._get_ds_model('d4')
        else:
            d_mod       = { ds : self._get_ds_model(ds) for ds in self._d_eff }

        for key, (mod_ee, mod_mm) in d_mod.items():
            self._plot_model(f'{key}_ee', mod_ee)
            self._plot_model(f'{key}_mm', mod_mm)
    
        self._d_mod = d_mod
    
        return self._d_mod
    #----------------------------------------------------
    @staticmethod
    def get_eff(kind='equal'):
        d_eff = {}
        if   kind == 'diff':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.5, 0.2)
            d_eff['d3'] = (0.7, 0.3)
            d_eff['d4'] = (0.8, 0.4)
        elif kind == 'half':
            d_eff['d1'] = (0.6, 0.3)
            d_eff['d2'] = (0.6, 0.3)
            d_eff['d3'] = (0.6, 0.3)
            d_eff['d4'] = (0.6, 0.3)
        elif kind == 'equal':
            d_eff['d1'] = (0.3, 0.3)
            d_eff['d2'] = (0.3, 0.3)
            d_eff['d3'] = (0.3, 0.3)
            d_eff['d4'] = (0.3, 0.3)
        elif kind == 'bias':
            d_eff['d1'] = (0.6, 0.25)
            d_eff['d2'] = (0.6, 0.25)
            d_eff['d3'] = (0.6, 0.25)
            d_eff['d4'] = (0.6, 0.25)
        else:
            log.error(f'Invalid kind: {kind}')
            raise
    
        return d_eff
#----------------------------------------------------




from np_reader import np_reader as np_rdr
import pprint
import pytest
import os

#-----------------------------
def skip_test():
    try:
        uname = os.environ['USER']
    except:
        pytest.skip()

    if uname in ['angelc', 'campoverde']:
        return

    pytest.skip()
#-----------------------------
def test_simple():
    skip_test()
    rdr       = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache = True
    d_rjpsi   = rdr.get_rjpsi()
    d_eff     = rdr.get_eff()
    cov_sys   = rdr.get_cov(kind='sys')
    cov_sta   = rdr.get_cov(kind='sta')
    d_byld    = rdr.get_byields()
#-----------------------------
def test_tarball():
    skip_test()
    rdr           = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache     = True
    rdr.cache_dir = 'tests/np_reader/tarball'
    d_rjpsi       = rdr.get_rjpsi()
    d_eff         = rdr.get_eff()
    cov_sys       = rdr.get_cov(kind='sys')
    cov_sta       = rdr.get_cov(kind='sta')
    d_byld        = rdr.get_byields()
#-----------------------------
def test_cache():
    skip_test()
    rdr           = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache     = False 
    rdr.cache_dir = 'tests/np_reader/cache'
    d_rjpsi       = rdr.get_rjpsi()
    d_eff         = rdr.get_eff()
    cov_sys       = rdr.get_cov(kind='sys')
    cov_sta       = rdr.get_cov(kind='sta')
    d_byld        = rdr.get_byields()
#-----------------------------
def main():
    test_simple()
    test_tarball()
    test_cache()
#-----------------------------
if __name__ == '__main__':
    main()




from setuptools import setup, find_packages

import glob

setup(
    name              = 'rk_extractor',
    version           = '0.1.6',
    description       = 'Used to extract RK from simultaneous fits',
    scripts           = glob.glob('scripts/jobs/*') + glob.glob('scripts/offline/*'),
    long_description  = '',
    package_dir       = {'' : 'src'},
    packages          = [''],
    install_requires  = open('requirements.txt').read().splitlines()
)




import numpy
import zfit
import matplotlib.pyplot as plt
import utils_noroot      as utnr

from logzero     import logger as log
from zutils.plot import plot   as zfp

#--------------------------------
class extractor:
    def __init__(self):
        self._d_eff       = None
        self._cov         = None
        self._d_rjpsi     = None
        self._d_data      = None
        self._d_model     = None
        self._d_ck        = None 
        self._l_dsname    = []

        self._rk          = zfit.Parameter('rk', 1.0, 0.0, 2.0)

        self._plt_dir     = None

        self._initialized = False
    #--------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._l_dsname = list(self._d_eff.keys())

        self._check_inputs()
        self._d_ck = { dsname : zfit.Parameter(f'ck_{dsname}', 1.0, 0.0, 2.0) for dsname in self._l_dsname }

        d_model = {}
        for ds, (mod_mm, mod_ee) in self._d_model.items():
            mod_ee = self._reparametrize_ee_pdf(ds, mod_mm, mod_ee)

            d_model[ds] = (mod_mm, mod_ee)

        self._d_model = d_model

        self._initialized = True
    #--------------------------------
    def _check_inputs(self):
        for ds, (eff_mm, eff_ee) in self._d_eff.items():
            self._check_eff(eff_mm)
            self._check_eff(eff_ee)

        for ds, (mod_mm, mod_ee) in self._d_model.items():
            self._check_model(mod_mm)
            self._check_model(mod_ee)

        for ds, (dat_mm, dat_ee) in self._d_data.items():
            self._check_data(dat_mm)
            self._check_data(dat_ee)

        self._check_keys(self._d_eff  , self._d_model)
        self._check_keys(self._d_model, self._d_data )
        self._check_keys(self._d_data , self._d_eff  )

        self._check_cov()
        self._check_rjpsi()
    #--------------------------------
    def _check_keys(self, d1, d2):
        if d1.keys() != d2.keys():
            log.error('Keys differ:')
            log.error(d1.keys())
            log.error(d2.keys())
            raise ValueError
    #--------------------------------
    def _check_cov(self):
        if self._cov is None:
            return

        if not isinstance(self._cov, numpy.ndarray):
            log.error(f'Covariance matrix is not a numpy array')
            raise

        ndset = len(self._d_eff)

        if self._cov.shape != (ndset, ndset):
            log.error(f'Covariance matrix has wrong shape: {shape._cov.shape}')
            raise
    #--------------------------------
    def _check_rjpsi(self):
        if self._d_rjpsi is None:
            self._d_rjpsi = {dsname : 1 for dsname in self._l_dsname}
            log.warning('rjpsi not found, using 1 for all datasets')
    #--------------------------------
    def _reparametrize_ee_pdf(self, dsname, mod_mm, mod_ee):
        nsg_mm  = self._get_yld(mod_mm, 'nsg_mm')
        nsg_ee  = self._get_yld(mod_ee, 'nsg_ee')

        l_model = mod_ee.models
        s_param = mod_ee.get_params(is_yield=True)
        l_param = list(s_param)
        index   = l_param.index(nsg_ee)

        eff_mm, eff_ee = self._d_eff[dsname]

        rjpsi = self._d_rjpsi[dsname]

        ck=self._d_ck[dsname]
        ck.set_value( (eff_ee / eff_mm) / rjpsi )

        if self._cov is None:
            ck.floating = False
            log.warning('Covariance matrix not specified, will fix ck in fit')

        nsg_ee = zfit.ComposedParameter(f'nsg_ee_rk_{dsname}', lambda a, b, c: b * c / a, params=[self._rk, nsg_mm, ck])

        log.debug(f'nsg_ee_rk_{dsname} -> {nsg_mm.name} * {ck.name} / {self._rk.name}')

        l_param[index] = nsg_ee

        l_model_ext  = [ model.copy().create_extended(nevt) for model, nevt in zip(l_model, l_param) ]

        mod_ee = zfit.pdf.SumPDF(l_model_ext)

        return mod_ee
    #--------------------------------
    def _get_yld(self, model, preffix):
        l_param   = model.get_params(is_yield=True)

        l_sig_yld = [ param for param in l_param if param.name.startswith(preffix)]
        [sig_yld] = l_sig_yld

        return sig_yld
    #--------------------------------
    @property
    def plt_dir(self):
        return self._plt_dir 

    @plt_dir.setter
    def plt_dir(self, value):
        try:
            self._plt_dir = utnr.make_dir_path(value)
        except:
            log.error(f'Cannot create: {value}')
            raise
    #--------------------------------
    @property
    def rjpsi(self):
        return self._d_rjpsi

    @rjpsi.setter
    def rjpsi(self, value):
        self._d_rjpsi = value
    #--------------------------------
    @property
    def eff(self):
        return self._d_eff

    @eff.setter
    def eff(self, value):
        self._d_eff = value
    #--------------------------------
    @property
    def cov(self):
        return self._cov

    @cov.setter
    def cov(self, value):
        self._cov = value
    #--------------------------------
    @property
    def data(self):
        return self._d_data

    @data.setter
    def data(self, value):
        self._d_data = value
    #--------------------------------
    @property
    def model(self):
        return self._d_model

    @model.setter
    def model(self, value):
        self._d_model = value
    #--------------------------------
    def _check_data(self, obj):
        if not isinstance(obj, zfit.data.Data):
            log.error(f'Object introduced is not a zfit dataset')
            raise ValueError
    #--------------------------------
    def _check_model(self, obj):
        if not isinstance(obj, zfit.pdf.SumPDF):
            log.error(f'Object introduced is not a zfit PDF')
            raise ValueError

        pdf = obj
        if not pdf.is_extended:
            log.error(f'PDF is not extended:')
            print(pdf)
            raise ValueError

        l_yld = pdf.get_params(is_yield=True)
        l_yld_name = [ yld.name for yld in l_yld if yld.name.startswith('nsg_') ]

        try:
            [yld_name] = l_yld_name
        except:
            log.error('Not found one and only one signal yield:')
            print(l_yld_name)
            raise ValueError

        log.debug(f'Picking up component with signal yield: {yld_name}')
    #--------------------------------
    def _check_eff(self, eff):
        if not isinstance(eff, float):
            log.error(f'Efficiency is not a float: {eff}')
            raise ValueError

        if not ( 0 < eff < 1 ):
            log.error(f'Efficiency is not in (0, 1): {eff:.3f}')
            raise ValueError
    #--------------------------------
    def _plot(self, data, model, results, component):
        if self._plt_dir is None:
            return

        obj=zfp(data=data, model=model, result=results)
        obj.plot(nbins=50) 

        plt_path = f'{self._plt_dir}/fit_{component}.png'
        log.info(f'Saving to: {plt_path}')
        plt.savefig(plt_path, bbox_inches='tight')
    #--------------------------------
    def _finalize(self):
        self._delete_par('rk')

        for ck in self._d_ck.values():
            self._delete_par(ck.name)

        for dsname in self._l_dsname:
            self._delete_par(f'nsg_ee_rk_{dsname}')
    #--------------------------------
    def _delete_par(self, par_name):
        if par_name in zfit.Parameter._existing_params:
            del zfit.Parameter._existing_params[par_name]
    #--------------------------------
    def _add_const(self, nll):
        if self._cov is None:
            log.warning(f'Covariance matrix is missing, not adding constraints to  fit')
            return nll

        l_ck_par=[]
        l_ck_val=[]
        for dsname in self._l_dsname:
            eff_mm, eff_ee = self._d_eff[dsname]
            rjpsi          = self._d_rjpsi[dsname]

            ck_val         = (eff_ee / eff_mm) / rjpsi
            ck_par         = self._d_ck[dsname]

            l_ck_val.append(ck_val)
            l_ck_par.append(ck_par)

        log.debug(f'Using covariance:\n{self._cov}')
        log.debug(f'Using ck:\n{l_ck_val}')
        cns_ck   = zfit.constraint.GaussianConstraint(params     = l_ck_par,
                                                      observation= l_ck_val,
                                                      uncertainty= self._cov)

        nll = nll.create_new(constraints=cns_ck)

        return nll
    #--------------------------------
    def _plot_fit(self):
        if self._plt_dir is None:
            return

        for dsname in self._l_dsname:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            self._plot(dat_mm, mod_mm, None, f'mm_{dsname}')
            self._plot(dat_ee, mod_ee, None, f'ee_{dsname}')
    #--------------------------------
    def get_fit_result(self):
        self._initialize()

        nll = None 
        log.info('Creating likelihoods')
        for dsname in self._l_dsname:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            nll_mm = zfit.loss.ExtendedUnbinnedNLL(model=mod_mm, data=dat_mm)
            nll_ee = zfit.loss.ExtendedUnbinnedNLL(model=mod_ee, data=dat_ee)

            if nll is None:
                nll = nll_mm + nll_ee
            else:
                nll+= nll_mm + nll_ee

        log.info('Adding constraints')
        nll = self._add_const(nll)

        mnm = zfit.minimize.Minuit()
        log.info('Running minimization')
        res = mnm.minimize(nll)

        self._plot_fit()

        log.info('Finalizing')
        self._finalize()

        return res 
#--------------------------------




import os
import numpy
import tarfile
import utils_noroot as utnr

from ndict             import ndict
from logzero           import logger         as log
from rk.eff_yld_loader import eff_yld_loader as eyl

#----------------------------------------
class np_reader:
    '''
    Class used to read nuisance parameters to calculate RK
    '''
    #------------------------
    def __init__(self, sys=None, sta=None, yld=None):
        '''
        sys (str): Version of efficiencies obtained when assessing systematics
        sta (str): Version of efficiencies obtained when assessing statistical uncertainties with bootstrapping
        yld (str): Version of fitted data yields (only Jpsi and Psi2S)
        '''
        self._sys         = sys 
        self._sta         = sta
        self._yld         = yld

        self._cov_dir     = None
        self._cache_dir   = None
        self._d_yld       = ndict()
        self._d_eff       = ndict()
        self._sys_flg     = 'pall_tall_gall_lall_hall_rall_qall_bnom_snom'
        self._l_ds_lab    = ['r1_TOS', 'r1_TIS', 'r2p1_TOS', 'r2p1_TIS', '2017_TOS', '2017_TIS', '2018_TOS', '2018_TIS']

        self._cache       = False
        self._initialized = False 
    #------------------------
    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, value):
        if value not in [True, False]:
            log.error(f'Invalid cache value: {value}')
            raise ValueError

        self._cache = value
    #------------------------
    @property
    def cache_dir(self):
        if self._cache_dir is None:
            return

        return cache_dir

    @cache_dir.setter
    def cache_dir(self, value):
        cache_dir = f'{value}/{self._sys}_{self._sta}_{self._yld}'
        try:
            os.makedirs(cache_dir, exist_ok=True)
        except:
            log.error(f'Cannot create: {cache_dir}')
            raise ValueError

        self._cache_dir = cache_dir
    #------------------------
    def _initialize(self):
        if self._initialized:
            return

        if self._cache_dir is None:
            self.cache_dir = '/tmp/np_reader'
            log.warning(f'Using default caching directory: {self._cache_dir}')

        self._cache_data(channel='ee')
        self._cache_data(channel='mm')
        self._check_data()
        self._tar_data()

        self._initialized = True
    #------------------------
    def _check_data(self):
        for key, l_yld in self._d_yld.items():
            if len(l_yld) == len(self._l_ds_lab):
                continue

            log.error(f'Lengths of yields for {key} does not align with labels: {self._l_ds_lab} <=> {l_yld}')
            raise

        for key, l_eff in self._d_eff.items():
            if len(l_eff) == len(self._l_ds_lab):
                continue

            log.error(f'Lengths of efficiencies for {key} does not align with labels: {self._l_ds_lab} <=> {l_eff}')
            raise
    #------------------------
    def _tar_data(self):
        tar_path = f'{self._cache_dir}.tar.gz'

        if os.path.isfile(tar_path):
            return

        json_dir = os.path.basename(self._cache_dir)
        with tarfile.open(tar_path, 'w:gz') as tar:
            tar.add(self._cache_dir, arcname=json_dir)

        log.info(f'Making: {tar_path}')
    #------------------------
    def _cache_data(self, channel=None):
        '''
        For a given channel, ee or mm, it will fill the dictionary of efficiencies for each
        dataset and trigger.
        '''
        if channel not in ['ee', 'mm']:
            log.error(f'Wrong channel: {channel}')
            raise ValueError

        eff_cache_path = f'{self._cache_dir}/eff_{channel}.json'
        yld_cache_path = f'{self._cache_dir}/yld_{channel}.json'
        sys_cache_path = f'{self._cache_dir}/sys_{channel}.json'
        sta_cache_path = f'{self._cache_dir}/sta_{channel}.json'

        is_cov = os.path.isfile(sys_cache_path) and os.path.isfile(sta_cache_path)
        is_eyl = os.path.isfile(eff_cache_path) and os.path.isfile(yld_cache_path) 

        if self._cache and is_eyl and is_cov:
            log.info(f'Picking cached data from: {eff_cache_path}')
            log.info(f'Picking cached data from: {yld_cache_path}')
            log.info(f'Picking cached data from: {sys_cache_path}')
            log.info(f'Picking cached data from: {sta_cache_path}')

            self._d_eff = utnr.load_json(eff_cache_path, decoder=ndict.json_decoder)
            self._d_yld = utnr.load_json(yld_cache_path, decoder=ndict.json_decoder)
            self._cvsys = utnr.load_json(sys_cache_path)
            self._cvsta = utnr.load_json(sta_cache_path)

            return

        l_trg = ['ETOS', 'GTIS'] if channel == 'ee' else ['MTOS', 'MTOS']
        for proc in ['sign', 'ctrl']:
            l_eff = []
            l_yld = []
            for year in ['r1', 'r2p1', '2017', '2018']:
                for trig in l_trg:
                    eff, yld = self._get_eff_yld(f'{proc}_{channel}', year, trig)

                    l_eff.append(eff)
                    l_yld.append(yld)

            self._d_eff[proc, channel] = l_eff
            self._d_yld[proc, channel] = l_yld

        eff_dir       = os.environ['EFFDIR']
        self._cov_dir = f'{eff_dir}/../covariance'
        self._cvsys   = self._get_cov(kind='sys')
        self._cvsta   = self._get_cov(kind='sta')

        utnr.dump_json(self._d_eff, eff_cache_path, encoder=ndict.json_encoder)
        utnr.dump_json(self._d_yld, yld_cache_path, encoder=ndict.json_encoder)
        utnr.dump_json(self._cvsys, sys_cache_path)
        utnr.dump_json(self._cvsta, sta_cache_path)
    #------------------------
    def _get_eff_yld(self, proc, year, trig):
        '''
        Will return numerical value of efficiency and fitted yield, for a specifi process
        in a year and trigger
        '''
        obj        = eyl(proc, trig, year, self._sys_flg)
        obj.eff_var= 'B_PT'
        t_yld, d_eff = obj.get_values(eff_version = self._sys, yld_version=self._yld)

        ctf  = d_eff['nom', 'B_PT']
        deff = ctf.efficiency
        oeff = deff.efficiency()
        eff  = oeff.val[0]
        yld  = t_yld[0]

        return eff, yld
    #------------------------
    def get_cov(self, kind=None):
        '''
        Will return covariance matrix (nxn numpy array)
        '''
        self._initialize()

        if kind not in ['sys', 'sta']:
            log.error(f'Invalid covariance kind: {kind}')
            raise ValueError

        cov_mat = self._cvsys if kind == 'sys' else self._cvsta

        return numpy.array(cov_mat)
    #------------------------
    def _get_cov(self, kind=None):
        if kind not in ['sys', 'sta']:
            log.error(f'Invalid uncertainty: {kind}')
            raise ValueError

        eff_ver  = self._sys if kind == 'sys' else self._sta
        pkl_path = f'{self._cov_dir}/{eff_ver}_{self._yld}/rx/matrix_abs_rc/tot.pkl'
        log.info(f'Picking up covariance from: {pkl_path}')
        cov      = utnr.load_pickle(pkl_path)

        return cov.tolist()
    #------------------------
    def get_eff(self):
        '''
        Will return rare mode efficiencies

        d_eff (dict): Dictionary {ds : (eff_mm, eff_ee)} with efficiency objects
        '''
        self._initialize()

        l_eff_rare_mm = self._d_eff['sign', 'mm'] 
        l_eff_rare_ee = self._d_eff['sign', 'ee'] 

        d_eff = {}
        for ds_lab, eff_mm, eff_ee in zip(self._l_ds_lab, l_eff_rare_mm, l_eff_rare_ee):
            d_eff[ds_lab] = eff_mm, eff_ee

        return d_eff 
    #------------------------
    def get_byields(self):
        '''
        Will return dictionary with efficiency corrected yields {ds : yld}
        e.g. {'r1_TIS_ee': 40021323}
        '''
        self._initialize()

        arr_eff_jpsi_mm = numpy.array(self._d_eff['ctrl', 'mm'])
        arr_eff_jpsi_ee = numpy.array(self._d_eff['ctrl', 'ee'])

        arr_yld_jpsi_mm = numpy.array(self._d_yld['ctrl', 'mm'])
        arr_yld_jpsi_ee = numpy.array(self._d_yld['ctrl', 'ee'])

        arr_yld_jpsi_mm = arr_yld_jpsi_mm / arr_eff_jpsi_mm
        arr_yld_jpsi_ee = arr_yld_jpsi_ee / arr_eff_jpsi_ee 

        d_byld_ee = {f'{ds}_ee' : byld for ds, byld in zip(self._l_ds_lab, arr_yld_jpsi_ee)}
        d_byld_mm = {f'{ds}_mm' : byld for ds, byld in zip(self._l_ds_lab, arr_yld_jpsi_mm)}

        d_byld= {}
        d_byld.update(d_byld_ee)
        d_byld.update(d_byld_mm)

        return d_byld
    #------------------------
    def get_rjpsi(self):
        '''
        Will return an array with rjpsi for every trigger and dataset 
        '''
        self._initialize()

        arr_eff_jpsi_mm = numpy.array(self._d_eff['ctrl', 'mm'])
        arr_eff_jpsi_ee = numpy.array(self._d_eff['ctrl', 'ee'])

        arr_yld_jpsi_mm = numpy.array(self._d_yld['ctrl', 'mm'])
        arr_yld_jpsi_ee = numpy.array(self._d_yld['ctrl', 'ee'])

        arr_rjpsi = (arr_yld_jpsi_mm / arr_yld_jpsi_ee) * (arr_eff_jpsi_ee / arr_eff_jpsi_mm)

        d_rjpsi   = {ds : rjpsi for ds, rjpsi in zip(self._l_ds_lab, arr_rjpsi)}

        return d_rjpsi
#----------------------------------------

